/***********************************************************************
 * Time-stamp: <Wednesday, 11.07.2001 at 20:23:58; edited by bauke>
 * 
 * Tina's random number generators
 *
 * testing equidistribution of bits of random number generators
 *
 * Copyright by Heiko Bauke, Magdeburg 2001 
 *
 * heiko.bauke@student.uni-magdeburg.de
 *
 * This is free software. You may use and distribute this software under
 * the terms of the GNU GENERAL PUBLIC LICENSE Version 2. See COPYING
 * for details.
 *
 ***********************************************************************/

#include <cstdlib>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <trng.h>

using namespace TRNG;
using namespace std;


template<class RNG_type>
void bit_equidist(RNG_type &r, long bit, long runs, long n) {
  // r: random number generator
  // bit: bit to test
  // runs: number of Chi-Square tests
  // n: random numbers per test
  vector<double> observ, prob;
  long i, num_fail, mask;
  double Chi2, Chi2P;
  cout.setf(ios::fixed);
  observ.resize(2, 0.0);
  prob.resize(2, 0.5);
  mask=1l<<bit;
  num_fail=0l;
  while (runs>0l) {
    observ[0]=0.0;
    observ[1]=0.0;
    for (i=0l; i<n; ++i)
      if ((r.rand()&mask)>0l)
	++observ[0];
      else
	++observ[1];
    Chi2P=0.0;
    try {
      Chi2=chi_square_test(prob, observ);
      Chi2P=chi_square_prob(Chi2, 1l);
    }
    catch (error err) {
      cerr << err.message << endl;
    }
    if (Chi2P<0.01 || 0.99<Chi2P) {
#ifdef LaTeX
      cout << "\\textbf{";
#endif
      ++num_fail;
    }
    cout << setfill(' ') << setprecision(3) << Chi2P;
#ifdef LaTeX
    if (Chi2P<0.01 || 0.99<Chi2P)
      cout << "}";
    cout << " & ";
#else
    cout << '\t';
#endif
    --runs;
  }
  if (num_fail>2l)
#ifdef LaTeX
    cout << "\\textbf{durchgefallen}";
#else
    cout << "failed";
#endif
  else
    if (num_fail>1l) 
#ifdef LaTeX
      cout << "\\textbf{suspekt}";
#else
      cout << "suspect";
#endif
    else
#ifdef LaTeX
      cout << "o.\\,k.";
#else
      cout << "o.k.";
#endif;
#ifdef LaTeX
  cout << "\\\\" << endl;
#else
  cout << endl;
#endif
}

template<class RNG_type>
void bit_equidist_main(RNG_type &r, long runs, long sub_max, long n) {
  long i, j;
#ifdef LaTeX
  cout << "\\multicol{" << runs+2l <<"}{@{}l@{}}{Generator " << r.name() 
       << "}\\\\" << endl << "\\midrule" << endl;
#else
  cout << "Generator " << r.name() << endl;
#endif
  for (j=1l; j<=sub_max; ++j) {
#ifdef LaTeX
    cout << "\\multicol{" << runs+2l <<"}{@{}l@{}}{Splitting level " << j 
	 << "}\\\\" << endl;
#else
    cout << "Splitting level " << j << endl;
#endif
    for (i=0l; i<31l; ++i) {
#ifdef LaTeX
      cout << "Bit Nr. " << i << " & ";
#else
      cout << "Bit Nr. " << i << '\t';
#endif
      r.reset();
      r.split(j, 0l);
      bit_equidist(r, i, runs, n);
    }
  }
}


int main(int argc, char *argv[]) {
  if (argc!=4) {
    cerr << 
      "Tina's Random Number Generator Library " << TRNG::version() << "\n";
    cerr << 
      "(P) & (C) by Heiko Bauke, Magdeburg 2001\n\n";
    cerr << 
      "Equdistribution test for bits\n";
    cerr << 
      "-----------------------------\n\n";
    cerr << 
      "synopsis:\n";
    cerr << 
      "$ " << argv[0] << " runs sub_max n\n\n";
    cerr <<
      "usefull arguments: runs=5; sub_max=128; n=3000;\n\n"; 
    cerr << 
      "Tests the equdistribution of the bits 0..30 and makes a Chi^2-test.\n\n";
    cerr 
      << "Four further informations see Donald E. Knuth\n";
    cerr << "\"The Art of Computer Programming\", 1981\n";
  } else {
    try {
      long runs, sub_max, n;
      runs=atol(argv[1]);
      sub_max=atol(argv[2]);
      n=atol(argv[3]);
#ifdef LaTeX
      cout << n << " W\\\"urfe je Test\\\\" << endl;
#else
      cout << n << " throws per test" << endl;
#endif
      ParkMiller r1;
      LCG32 r2;
      LCG64 r3;
      MRG2 r4;
      MRG3 r5;
      MRG4 r6;
      CLCG2 r7;
      CLCG3 r8;
      CLCG4 r9;
      EINV r10;
      EINVLCG64 r11;
      bit_equidist_main(r1, runs, sub_max, n);
      bit_equidist_main(r2, runs, sub_max, n);
      bit_equidist_main(r3, runs, sub_max, n);
      bit_equidist_main(r4, runs, sub_max, n);
      bit_equidist_main(r5, runs, sub_max, n);
      bit_equidist_main(r6, runs, sub_max, n);
      bit_equidist_main(r7, runs, sub_max, n);
      bit_equidist_main(r8, runs, sub_max, n);
      bit_equidist_main(r9, runs, sub_max, n);
      bit_equidist_main(r10, runs, sub_max, n);
      bit_equidist_main(r11, runs, sub_max, n);
    }
    catch (...) {
      cerr << "something went wrong" << endl;
    }
  }
  return EXIT_SUCCESS;
}
